
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, shutil, textwrap, warnings, time, md5
import numarray
from numarray.numerictypes import *
import Image
import utilities, pathnames, scorearr, constants
from constants import *
from LN import LN, MLLN

def NULL(*args, **kwargs):
    pass
#ANNOUNCE = MLLN
ANNOUNCE = NULL

# Verbal shortcut: "an ImData object" or an "ImData"  means "an instance of
# Imdata". The variable "ID" is often used for an ImData instance. Also
# "N" for a Name object.
# "iff" means "if and only if".

#>=== Name ====================================================

class Name(str):
    def __new__(cls, name, mtime, topdirs):
        """Wrapper around a pathname."""
        # When unpickling takes place, there may be no file named "name".
        # Therefore I cannot find mtime here.
        name = os.path.abspath(name)
        self = str.__new__(cls, name)
        self.mtime = mtime
        self.topdirs = topdirs
        self.make_virtuals()
        return self

    # Needed for unpickling.
    def __getnewargs__(self):
        return (str.__str__(self), self.mtime, self.topdirs)

    def __str__(self):
        template = """\
            <Name
                %s
                mtime: %f>\n"""
        template = textwrap.dedent(template)
        return template % (str.__str__(self), self.mtime)

    def __repr__(self):
        template = """\
            <Name
                %s
                mtime: %f,
                vn: %s
                vd: %s
                topdirs=%s>\n"""
        template = textwrap.dedent(template)
        text = textwrap.fill(str(self.topdirs), subsequent_indent='  ')
        return template % (str.__repr__(self), self.mtime,
                self.virtual_name, self.virtual_dirname(), text)

    def as_str(self):
        return str.__str__(self)

    def make_virtuals(self):
        # The virtuals are used in DDL.py and html.py. They are strings, not
        # Names. The topdirs are assumed to have been normalized by running
        # them through PathNames.
        for topdir in self.topdirs:
            if pathnames.isunder(self, topdir):
                break
        else:
            raise RuntimeError('BUG: name not under any topdir')
        firstpart = len(os.path.dirname(topdir))
        self.virtual_name = str(self[firstpart:])
        # self.virtual_components = pathnames.splitall(self.virtual_name)
        self.level = self.virtual_name.count('/')

    def virtual_dirname(self):
        return os.path.dirname(self.virtual_name[1:])

#>=== ImData ==================================================

class ImData(list):
    """Stores information about a image file"""
    def __init__(self, N=None, topdirs=None):
        """An ImData instance is a list of Name's."""
        if N is None != topdirs is None:
            raise TypeError('N and topdir must be both present or neither present')
        list.__init__(self)
        self.hashcode = None  # md5 hash of the image. Assigned by
                              # get_image_data.
        self.imnum = None  # Image number. Assigned by MRQuery.
        self.deleted = False  # True if this ImData instance is being cached
                              # for possible reuse.
        self.bwbit = None  # True for black & white images.
        self.cols, self.rows = None, None
        self.bytes = None     # Size of image file.
        if N is not None:
            self.get_image_data(N, topdirs)

    def get_image_data(self, N, topdirs):
        """Gets data from an image"""
        if N in self:
            raise TypeError('cannot add a Name that is already present')
        if self.hashcode is not None:
            self.append(N)
            return

        # Get information from the image. Use md5 to calculate a hashcode
        # for the pixels of the image. If two different images have the same
        # md5, then two differing images will be treated as identical.  If
        # this _rare_ event is a problem for you, use MRQuery.paranoid().  In
        # the future, I might relax the definition of duplicate images to
        #    score(name1, name2) >= C
        # where C might be .9.
        im = Image.open(N)
        try:
            s = im.tostring()
        except Exception:
            print >> sys.stderr, 'Problem with', N.as_str()
            raise
        self.hashcode = utilities.Hash(md5.new(s).digest())
        self.append(N)
        if im.mode not in ('L', 'P', 'RGB'):
            raise TypeError("image mode must be 'L', 'P', or 'RGB'")
        self.bwbit = (im.mode == 'L')
        self.cols, self.rows = im.size
        self.bytes = os.path.getsize(N)

    def __str__(self):
        template = """\
            <ImData: 
            imnum: %s
            hashcode: %s
            deleted: %s
            %s>\n"""
        template = textwrap.dedent(template)
        out = []
        for N in self:
            out.append(textwrap.fill(str(N), subsequent_indent='    '))
        text = '\n'.join(out)
        text = template % (self.imnum, self.hashcode, self.deleted, text)
        return utilities.redent(text, 0, 4)

    def __repr__(self):
        template = """\
            <ImData:
                hashcode: %s
                imnum: %s
                deleted: %s
                bwbit: %s
                rows: %s, cols: %s>
                %s\n"""
        template = textwrap.dedent(template)
        out = []
        for N in self:
            out.append(textwrap.fill(str(N), subsequent_indent='    '))
        text = '\n'.join(out)
        return template % (self.hashcode, self.imnum, self.deleted, self.bwbit,
                 self.rows, self.cols, text)

    def tonames(self):
        out = list()
        for N in self:
            out.append(N.as_str())
        return tuple(out)

    def check_invariants(self):
        if not constants.CHECK_INVARIANTS:
            return
        for N in self:
            if not isinstance(N, Name):
                raise TypeError('non-Name object in this ImData instance')
        if len(set(self)) != len(self):
            raise TypeError('pathname occurs in two different Names')

#>=== MRQuery =======================================================

class MRQuery(object):
    """a collection of ImData's"""
    def __init__(self, topdatadir, basename, topdirs=None, excluded_dirs=None):
        """Create, populate and store a MRQuery object.

        An image is in some instance of MRQuery iff it is under one of the
        topdirs.

        WARNING: Do not use filenames containing a space or other weird
                 character. This will expose several bugs in the code.

        "MRQuery" has a number of attributes.

        1. "filenames" contains a bunch of filenames.

        2. There are three dictionaries in a instance of MRQuery. The main
        one is "hash2ImData". The keys here are md5 hash codes. The values
        are instances of ImData. Since several images may be identical to
        each other, there may be several names associated with an instance
        of ImData. The inverse dict of "hash2ImData" is "name2hash". The
        keys of "name2hash" are pathnames; the values are hashes.

        3. For each hashcode there is a unique and permanent "imnum". The
        imnums are assigned as consecutive integers to each new hashcode.
        The imnums are the keys of "imnum2hash". There is no dictionary
        "hash2imnum". Use "self.hash2ImData[hashcode].imnum".

        4. "imnum_count" is the number of imnums currently in use and the
        next imnum to be assigned. Similarly "haar_count" is the number of
        haar transforms that have been done and the next imnum to be
        transformed.

        5. "Darrs" is a numarray used to store the results of haar
        transforms.  it is indexed by imnums. In "mrquery.pdf" tha results
        of the haar transforms are kept in a data structure "D".

        6. A bit in "bwbits" say whether an image is black and white or
        color.

        7. "killed_imnums" is a set of imnums being removed from the system.

        8. "sortlist" is a sorted list of scores found by comparing two
        images with each other.

        9. "ddl_sortlist" is "sortlist" with some elements removed because
        of "distance, direction, level" constraints. See "DDL.py".

        10. "block"s are used to improve memory localization in
        "scorearr.py".

        11. If "usewts" is True, the haar coefficients will be weighted
        during scoring.

        12. "self.html_ignores" is used by "html.py". If a pair of
        directories are both under a directory in "self.html_ignores",
        then no html is produced for the pair.

        13. "self.alldirs" contains a sorted list of those directories below
        the topdirs which contain images.

        When ImData's are added, removed or modified, changes need to be
        made in some of these attributes. When new images are added, the
        entire haar, scorearr, DDL, html process needs to be called. If an
        ImData object is removed, the scores for that image must be removed
        from "sortlist". When any ImData's are removed or modified,
        "ddl_sortlist" becomes invalid.
        """
        # The following must be called as early as possible when running the
        # mrquery family of programs.
        os.stat_float_times(True)
        self.filenames = pathnames.PathNames(topdatadir, basename, topdirs,
                           excluded_dirs)
        self.filenames.makedirs()  # Creates some directories under topdatadir.
        if os.path.exists(self.filenames.savefile):
            raise IOError('creating a new MRQuery object but a savefile exists')
        self.emptydirs()
        self.hash2ImData = dict()
        self.name2hash = dict()
        self.imnum_count = 0
        # Each hashcode will have a associated imnum. If a hashcode is
        # removed, its imnum is also removed.
        self.imnum2hash = dict()
        self.haar_count = 0
        self.Darrs = numarray.zeros( (LAYERS, 0, RCQUOTA) , Int32)
        self.bwbits = numarray.zeros( (0,), Bool)
        self.killed_imnums = set()
        self.resetScorearr()
        self.ddl_sortlist = list()
        self.html_ignores = list()
        self.alldirs = list()
        self.populate()

    def __str__(self):
        out = ['<MRQuery:']
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            out.append(str(ID))
        out.append('>')
        return '\n'.join(out)

    def __repr__(self):
        out = ['<MRQuery:']
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            out.append(repr(ID))
        out.append('>')
        return '\n'.join(out)

    def Store(self):
        """Stores self to disk"""
        self.check_invariants()
        if os.path.isfile(self.filenames.savefile):
            shutil.copyfile(self.filenames.savefile, self.filenames.bakfile)
        utilities.Store(self, self.filenames.savefile)

    def data_check(self, topdatadir, basename):
        """Does data read from disk have the correct topdatadir and basename?"""
        filenames = pathnames.PathNames(topdatadir, basename)
        if filenames.topdatadir != self.filenames.topdatadir or \
                filenames.basename != self.filenames.basename:
            template = """\
                topdatadirs self: %s, disk: %s
                basenames self: %s, disk: %s
                The savefile contains the wrong set of images."""
            template = textwrap.dedent(template)
            text = template % (filenames.topdatadir,
                   self.filenames.topdatadir, filenames.basename,
                   self.filenames.basename)
            raise IOError(text)

    def check_invariants(self):
        """Check required invariants in the four dicts in MRQuery.

        Any exception raised here mean that there is a BUG. The code is over
        documented to explain the invariants.

        These checks are slow. Some even run in O(N^2) time.

        Here is a list of the requirements that are tested.

            "os.stat_float_times(True)" must called.
            "hashcode" must be an instance of Hash.
            The values of self.hash2ImData are ImData instances.
            A deleted ImData instance has no names in it.
            ImData instance contains a non-Name object
            ID.imnum not an Imnum where ID is an ImData instance
            ID.imnum not in self.imnum2hash where ID is an ImData instance
            Each name occurs in only one ImData instance.
            Each ImData instance has a unique imnum.
            Each ImData instance has a unique hashcode.
            Each "hashcode" must be a Hash instance.
            The keys of "name2hash" are Name's.
            Each "imnum" must be an Imnum.

            The three dictionaries are "hash2ImData", "name2hash" and
            "imnum2hash". There is no "hash2imnum". Use
            "self.hash2ImData[hashcode].imnum".

            The keys of "hash2ImData" are hashcodes.
            The keys of "hash2ImData" are the same as the set of hashcodes
                      in the ImData instances.
            The values of "hash2ImData" are ImData instances.
            "hash2ImData" is one to one.

            The keys of "name2hash" are Name instances.
            The values of "name2hash" are hashcodes.
            The keys are "name2hash" are the same as the set of all the
                      Names in any of the ImData instances.
            A key in "hash2ImData" must also be a value of name2hash
                      unless hash2ImData[key] is deleted.

            The keys of "imnum2hash" are imnums.
            The values of "imnum2hash" are hashcodes.
            "imnum2hash" is one to one.
            "self.hash2ImData[hashcode].imnum" is the inverse of "imnum2hash".
            The keys of "imnum2hash" are the imnums that are in use.
            The values nof "imnum2hash" are the set of hashcodes in all the
                      ImData instances except those that have been
                      "deleted". The "deleted" ImData instances have no
                      names.
        """
        if not CHECK_INVARIANTS:
            return
        # Was "os.stat_float_times(True)" called?
        if not os.stat_float_times():
            raise TypeError('BUG: "os.stat_float_times(True)" has not been called')

        for hashcode in self.hash2ImData:
            if not isinstance(hashcode, utilities.Hash):
                raise TypeError('BUG: hashcode is not an instance of Hash. '
                     'It has type %s.' % type(hashcode))
            ID = self.hash2ImData[hashcode]
            if not isinstance(ID, ImData):
                raise TypeError(
                  'BUG: value in self.hash2ImData is not an ImData instance')
            if ID.deleted:
                if len(ID) != 0:
                    raise TypeError('BUG: ImData instance is deleted but there are still '
                         'names in it')
                continue
            for N in ID:
                if not isinstance(N, Name):
                    raise TypeError('BUG: ImData instance contains a non-Name object')
            if not isinstance(ID.imnum, utilities.Imnum):
                raise TypeError('BUG: ID.imnum not an Imnum where '
                     'ID is an ImData instance')
            if ID.imnum not in self.imnum2hash:
                # Each imnum must be a key of "self.imnum2hash".
                raise TypeError('BUG: ID.imnum not in self.imnum2hash '
                   'where ID is an ImData instance')

        names_in_IDs = list()
        imnums_in_IDs = list()
        hashes_in_IDs = list()
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            imnums_in_IDs.append(ID.imnum)
            hashes_in_IDs.append(ID.hashcode)
            for N in ID:
                names_in_IDs.append(N)
        if len(names_in_IDs) != len(set(names_in_IDs)):
            # for N in names_in_IDs:
            #     LN(N.as_str())
            raise TypeError("BUG: a name occurs twice in the ImData's")
        if len(imnums_in_IDs) != len(set(imnums_in_IDs)):
            raise TypeError("BUG: two ImData's have the same imnum")
        if len(hashes_in_IDs) != len(set(hashes_in_IDs)):
            raise TypeError("BUG: two ImData's have the same hashcode")

        h2ID_keys = self.hash2ImData.keys()
        h2ID_values = self.hash2ImData.values()
        n2h_keys = self.name2hash.keys()
        n2h_values = self.name2hash.values()
        i2h_keys = self.imnum2hash.keys()
        i2h_values = self.imnum2hash.values()

        if constants.QUADRATIC_CHECK_INVARIANTS:
            # This runs in time quadratic in the number of hashcodes.
            for i in range(len(h2ID_keys) - 1):
                for j in range(i+1, len(h2ID_keys)):
                    hashcodei = h2ID_keys[i]
                    hashcodej = h2ID_keys[j]
                    IMi = self.hash2ImData[hashcodei]
                    IMj = self.hash2ImData[hashcodej]
                    if not (IMi.deleted and IMj.deleted) and set(IMi) == set(IMj):
                        LN(i, self.hash2ImData[hashcodei])
                        LN(j, self.hash2ImData[hashcodej])
                        raise TypeError("BUG: self.hash2ImData is not one to one")
            if len(i2h_values) != len(set(i2h_values)):
                raise TypeError("BUG: self.imnum2hash is not one to one")

        names_in_IDs = set(names_in_IDs)
        imnums_in_IDs = set(imnums_in_IDs)
        hashes_in_IDs = set(hashes_in_IDs)
        h2ID_keys = set(h2ID_keys)
        n2h_keys = set(n2h_keys)
        n2h_values = set(n2h_values)
        i2h_keys = set(i2h_keys)
        i2h_values = set(i2h_values)

        for i, aset in enumerate( (h2ID_keys, n2h_values, i2h_values,
                   hashes_in_IDs) ):
            for hashcode in aset:
                if not isinstance(hashcode, utilities.Hash):
                    LN(i, type(hashcode))
                    LN(hashcode)
                    raise TypeError("BUG: hashcode not a Hash")
        for i, aset in enumerate( (n2h_keys, names_in_IDs) ):
            for N in aset:
                if not isinstance(N, Name):
                    LN(i, N)
                    raise TypeError("BUG: non-Name object in name2hash.keys()")
        for i, aset in enumerate( (i2h_keys, imnums_in_IDs) ):
            for imnum in aset:
                if not isinstance(imnum, utilities.Imnum):
                    LN(i, type(imnum), imnum)
                    LN(aset)
                    raise TypeError("BUG: imnum is not an Imnum")

        # Compare within three equivalence classes:
        #  1: (h2ID_keys, n2h_values, i2h_values, hashes_in_IDs)
        if h2ID_keys != n2h_values:
            # If ID.deleted is True, then ID.name is None and
            # len(ID.duplicates) == 0. Therefore there are some hashes in
            # hash2ImData that have no associated nmaes.
            LN(h2ID_keys <= n2h_values, h2ID_keys >= n2h_values,
                len(h2ID_keys - n2h_values), len(n2h_values - h2ID_keys))
            for hashcode in h2ID_keys - n2h_values:
                ID = self.hash2ImData[hashcode]
                if not ID.deleted:
                    LN(ID)
                    raise TypeError("BUG: A key in hash2ImData must also be a "
                       "value of name2hash unless hash2ImData[key] is deleted.")
            if len(n2h_values - h2ID_keys) > 0:
                 raise TypeError("BUG: A value of name2hash is not a key "
                      "in hash2ImData")
        if h2ID_keys != i2h_values:
            raise TypeError("BUG: set(self.hash2ImData.keys()) != "
                   "set(self.name2hash.values())")
        if h2ID_keys != hashes_in_IDs:
            LN(len(h2ID_keys), len(hashes_in_IDs))
            for hashcode in h2ID_keys - hashes_in_IDs:
                 LN(hashcode)
                 ID = self.hash2ImData[hashcode]
                 LN(ID)
            raise TypeError("BUG: set(self.hash2ImData.keys()) != "
                   "set(hashes_in_IDs)")

        #  2: (n2h_keys, names_in_IDs)
        if n2h_keys != names_in_IDs:
            diff = n2h_keys - names_in_IDs
            if len(diff) > 0:
                for N in diff:
                    LN(N)
                raise TypeError("BUG: extra Name's in self.name2hash")
            diff = names_in_IDs - n2h_keys
            if len(diff) > 0:
                MLLN(*names_in_IDs)
                MLLN(*n2h_keys)
                MLLN(*diff)
                raise TypeError("BUG: there is a name in an ImData which "
                     "is not in self.name2hash")

        #  3 : (i2h_keys, imnums_in_IDs)
        if i2h_keys != imnums_in_IDs:
           raise TypeError("BUG: set(self.imnum2hash.keys()) != "
                   "set(imnums_in_IDs)")

        # Check inverse dicts:
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            if not ID.deleted:
                for N in ID:
                    if N not in self.name2hash:
                        # Every N must be a key of "self.name2hash".
                        LN(ID)
                        raise TypeError('BUG: a Name is not in self.name2hash')
                    elif self.name2hash[N] != hashcode:
                        # Consistency check for "self.name2hash[name]".
                        raise TypeError('BUG: self.name2hash inconsistent '
                             'with self.hash2ImData')
        for N in self.name2hash:
            hashcode = self.name2hash[N]
            if N not in self.hash2ImData[hashcode]:
                # Every name must be in an ImData.
                raise TypeError("BUG: a key of self.name2hash is not in the "
                      "corresponding ImData: %s" % name)
        for imnum in self.imnum2hash:
            hashcode = self.imnum2hash[imnum]
            if self.hash2ImData[hashcode].imnum != imnum:
                # Consistency check for "self.imnum2hash".
                raise TypeError('BUG: self.hash2ImData[hashcode].imnum != imnum')

    def set_deleted(self, ID):
        """Convert an ImData instance to a "deleted" one.

        If another image with the same hashcode is found the "deleted"
        ImData object can be reused.
        """
        assert len(ID) == 1
        ANNOUNCE('ANNOUNCE: A ImData is being deleted', ID)
        if ID[0] in self.name2hash:
            del self.name2hash[ID[0]]
        # These are still True:
        # ID.hashcode in self.hash2ImData
        # ID.imnum in self.imnum2hash
        del ID[0]
        ID.deleted = True

    def remove(self, N):
        """Remove a Name instance from this MRQuery instance"""
        if not isinstance(N, Name):
            raise TypeError('"remove" called with a non-Name object')
        hashcode = self.name2hash[N]
        ID = self.hash2ImData[hashcode]
        if len(ID) > 1:
            # Keep ID but remove N.
            del ID[ID.index(N)]
            del self.name2hash[N]
        else:
            # ID contains no Names. Set to deleted.
            # Only self.clean() removes these.
            self.set_deleted(ID)
        ANNOUNCE('ANNOUNCE: removed N', N)

    def insert(self, Nnew):
        """Puts an Name instance into a MRQuery instance.

        In some cases, an ImData instance will be put into this MRQuery
        instance. In other cases, the data from an ImData instance will be
        put into some ImData instance that is already in self. This is done
        because the ImData instance in self may have already completed haar
        or scorearr.  This method would usually be called "set". But the
        semantics here are somewhat different so I have changed the name.

        This is the _only_ place where new ImData's are created.

        We assume that same pathname and same mtime implies same hashcode.

        This function is tricky or perhaps even delicate. Look for bugs here.
        """
        # This is the most common case.
        if Nnew in self.name2hash:
            hashcode = self.name2hash[Nnew]
            ID = self.hash2ImData[hashcode]
            i = ID.index(Nnew)
            if ID[i].mtime == Nnew.mtime:
                return

        IDnew = ImData(Nnew, self.filenames.topdirs)  # Moderately expensive.
        if Nnew in self.name2hash and \
                 self.name2hash[Nnew] != IDnew.hashcode:
            # Nnew is already in use for a different image. Remove the
            # current use of Nnew.
            self.remove(Nnew)
            ANNOUNCE('ANNOUNCE: removed current use of Name', Nnew)
        if IDnew.hashcode in self.hash2ImData:
            # An identical image is already in self.
            IDold = self.hash2ImData[IDnew.hashcode]
            if IDold.deleted:
                # Fix up the "deleted" ImData object and reuse it.
                IDold.deleted = False
                IDold.append(Nnew)
                # No change needed to hash2ImData or imnum2hash.
                self.name2hash[Nnew] = IDold.hashcode
                ANNOUNCE('ANNOUNCE: reuse deleted ImData', IDold)
            else:
                for N in IDold:
                    if N == Nnew and N.mtime != Nnew.mtime:
                        # The mtime was changed. Correct it.
                        N.mtime = Nnew.mtime
                        ANNOUNCE('ANNOUNCE: mtime needed to be changed')
                        return
                    # N == Nnew and N.mtime == Nnew.mtime is the most common
                    # case. See above.
                # Add new identical image.
                IDold.append(Nnew)
                self.name2hash[Nnew] = IDnew.hashcode
                ANNOUNCE('ANNOUNCE: name added to IDold', IDold)
        else:
            # Add a new image.
            self.hash2ImData[IDnew.hashcode] = IDnew
            self.name2hash[Nnew] = IDnew.hashcode
            an_imnum = utilities.Imnum(self.imnum_count)
            self.imnum2hash[an_imnum] = IDnew.hashcode
            IDnew.imnum = an_imnum
            self.imnum_count += 1
            ANNOUNCE('ANNOUNCE: added an image', IDnew)

    def make_alldirs(self):
        self.alldirs = [dirname for dirname in
                   utilities.WalkDirs(self.filenames.topdirs)]

    def populate(self):
        """Add new images to this MRQuery instance"""
        allnames = list()
        counter = 0
        t1 = t0 = t = time.time()
        for topdir in self.filenames.topdirs:
            for pathname in utilities.WalkImages(topdir,
                       excluded_dirs=self.filenames.excluded_dirs):
                # This is the _only_ place where Name's are created.
                # grep for mtime to see why this next line is here.
                mtime = os.path.getmtime(pathname)
                N = Name(pathname, mtime, self.filenames.topdirs)
                # This is the _only_ place where "insert" is called.
                self.insert(N)
                allnames.append(N)
                counter += 1
                if time.time() - t1 >= constants.SAVETIME:
                    ta = time.time()
                    self.Store()
                    now = time.time()
                    print 'Stored at %4.2f secs. Store took %4.2f secs.' % \
                       (now - t0, now - ta)
                    t1 = time.time()
                if counter % constants.MRQPRINTAT == 0:
                    tt = time.time()
                    print '%i images processed. This time: %4.2f, total time: %4.2f' % \
                            (counter, tt - t, tt-t0)
                    t = time.time()
        self.make_alldirs()
        self.Store()
        print counter, "images processed."
        LN(len(allnames), 'names after populating')
        return allnames

    def update(self):
        """Update this MRQuery object by adding and/or removing images"""
        oldnames = set(self.name2hash.keys())
        LN(len(oldnames), 'before updating')
        allnames = self.populate()
        allnames = set(allnames)
        # Find names that need deleting using set difference.
        oldnames -= allnames
        # LN(len(oldnames), 'Names need to be deleted.')

        self.check_invariants()
        for name in oldnames:
            self.remove(name)
        self.clean()
        # Remove deleted images from sortlist.
        self.update_sortlist()
        # Modifying MRQuery invalidates ddl_sortlist.
        self.ddl_list = list()
        self.Store()

    def clean(self):
        """Removes deleted ImData's from MRQuery.

        This needs to be done periodically. Otherwise "deleted" ID's are
        never removed. A good time to call clean is at the end of each 
        update.
        """
        kill = list()
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            # assert ID.hashcode == hashcode
            if ID.deleted:
                # assert len(ID.duplicates) == 0
                # assert ID.name is None
                kill.append(hashcode)
        for hashcode in kill:
            # The hashcode, the imnum and the ImData are completely removed.
            imnum = self.hash2ImData[hashcode].imnum
            del self.hash2ImData[hashcode]
            self.killed_imnums.add(imnum)
            del self.imnum2hash[imnum]

    def emptydirs(self):
        """Remove all the data files"""
        for adir in (self.filenames.datadir, self.filenames.htmldir,
                 self.filenames.thumbdir):
            if os.path.exists(adir):
                for name in os.listdir(adir):
                    fullname = os.path.join(adir, name)
                    if os.path.isdir(fullname):
                        shutil.rmtree(fullname)
                    else:
                        os.remove(fullname)

    def paranoid(self):
        """Finds any md5 hash collisions that occurred. Very rare."""
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            if len(ID.duplicates) == 0:
                continue
            s = open(ID.name, 'r').read()
            for name in ID.duplicates:
                if s != open(name, 'r').read():
                    message = 'md5 hash collision: %s, %s' % (ID.name, name)
                    warnings.warn(message)

    def relocate(self, topdatadir):
        """Use this if "_data" files have been moved to another topdatadir"""
        # Not debugged.
        self.filenames = pathnames.PathNames(topdatadir,
                 self.filenames.basename, self.filenames.topdirs)
        self.filenames.nakedirs()

    def hash_count(self):
        """How many hashcodes are in use?"""
        return len(self.hash2ImData.keys()) 

    def resetScorearr(self, usewts=None):
        """Reset the Scorearr data structures"""
        self.sortlist = list()
        self.ddl_sortlist = list()
        self.block = scorearr.Block(self.haar_count)
        self.usewts = usewts

    def haar_finished(self):
        """Have all the Haar transforms been done?"""
        return self.haar_count == self.imnum_count

    def bad_imnums(self):
        """Count the number of imnums that are no longer in use"""
        bad = numarray.zeros( (self.haar_count,), Bool) 
        for imnum in range(self.haar_count):
            bad[imnum] = (imnum not in self.imnum2hash)
        return bad

    def update_sortlist(self):
        """Update the sortlist by removing killed_imnums"""
        new_sortlist = list()
        for score, imnum1, imnum2 in self.sortlist:
            if imnum1 not in self.killed_imnums and \
                     imnum2 not in self.killed_imnums:
                new_sortlist.append( (score, imnum1, imnum2) )
        self.killed_imnums = set()
        self.sortlist = new_sortlist

    def replace_topdirs(self, newtopdirs):
        """Replace topdirs. An update must done after this is called."""
        newtopdirs = pathnames.prepare_topdirs(newtopdirs)
        # I think the program works no matter what self.filenames.topdirs
        # and newtopdirs are. The TypeError is for safety.
        ntds = set(newtopdirs)
        tds = set(self.filenames.topdirs)
        if not tds <= ntds:
            raise TypeError('self.filenames.topdirs is not a subset '
                    'of newtopdirs')
        self.filenames.topdirs = newtopdirs

    def replace_excluded_dirs(self, new_excluded_dirs):
        """Replace excluded_dirs. An update must done after this is called."""
        new_excluded_dirs = pathnames.prepare_excludes(new_excluded_dirs)
        # I think the program works no matter what
        # self.filenames.excluded_dirs and new_excluded_dirs are. The
        # TypeError is for safety.
        if not new_excluded_dirs <= self.filenames.excluded_dirs:
            raise TypeError('"new_excluded_dirs" is not a subset of '
                  'self.filenames.excluded_dirs')
        self.filenames.excluded_dirs = new_excluded_dirs

    def unexclude_dirs(self, unexcluded_dirs):
        """Remove some of the excluded_dirs. An update must done after this is called."""
        unexcluded_dirs = pathnames.prepare_excludes(unexcluded_dirs)
        # I think the program works works no matter what
        # self.filenames.excluded_dirs and unexcluded_dirs are. The TypeError is
        # for safety.
        ex = set(self.filenames.excluded_dirs)
        ex2 = set(unexcluded_dirs)
        # LN(self.filenames.excluded_dirs)
        if not ex2 <= ex:
            LN(sorted(ex2 - ex))
            raise TypeError('"unexcluded_dirs" is not a subset of '
                  'self.filenames.excluded_dirs')
        ex -= ex2
        self.filenames.excluded_dirs = tuple(ex)

    def output_excludes(self, filename):
        """Print the excluded_dirs to a file"""
        out = list(self.filenames.excluded_dirs)
        out.sort()
        out.append('')
        text = '\n'.join(out)
        open(filename, 'w').write(text)

    def countBW(self):
        """Count the color and B&W images by directory"""
        countBWdict = dict()
        for hashcode in self.hash2ImData:
            ID = self.hash2ImData[hashcode]
            for N in ID:
                Ndir = os.path.dirname(N)
                if Ndir not in countBWdict:
                    countBWdict[Ndir] = [0, 0]
                countBWdict[Ndir][1] += 1
                if ID.bwbit:
                    countBWdict[Ndir][0] += 1
        return countBWdict

#>=== Do, Fetch, and Retrieve ====================================

def Retrieve_MRQuery(topdatadir, basename):
    """Retrieve from disk a MRQuery instance"""
    # The following must be called as early as possible when running the
    # mrquery family of programs.
    os.stat_float_times(True)
    topdatadir, basename, savefile = \
               pathnames.savefilename(topdatadir, basename)
    mrq = utilities.Retrieve(savefile)
    # Backwards compatability.
    if not hasattr(mrq, 'html_ignores'):
         LN()
         mrq.html_ignores = list()
    mrq.data_check(topdatadir, basename)
    mrq.check_invariants()
    return mrq

def FetchOrCreate_MRQuery(topdatadir, basename, reuse_MRQuery, topdirs=None,
           excluded_dirs=None):
    """Create a MRQuery instance or read it from a file.

    If the instance is created, it will contain data about all the images
    under the topdirs except that it will not contain data about any images
    under the excluded_dirs.

    "reuse_MRQuery" can be 'new', 'update' or 'as_is':
       'new'    A new MRQuery instance is created. "topdirs" must be given.
                if "excluded_dirs" is None, it defaults to tuple().
       'update' A MRQuery instance is read from a file. It is updated to
                match the images under the topdirs. If some "topdirs" are
                given, the topdirs in the MRQuery instance are replaced. If
                some "excluded_dirs" are given, they are **REMOVED FROM** the
                excluded_dirs in the instance.
       'as_is'  A MRQuery instance is read from a file. No updates or
                changes are made. "topdirs" and "excluded_dirs" must be None.
    """
    if reuse_MRQuery not in ('new', 'update', 'as_is'):
        raise ValueError("reuse_MRQuery must be 'new', 'update', or 'as_is'")
    a = topdirs is None
    b = excluded_dirs is None
    if reuse_MRQuery == 'as_is' and (not a or not b):
        raise TypeError("if reuse_MRQuery is "
         "'as_is', then topdirs and excluded_dirs must be None")
    if reuse_MRQuery == 'new' and a:
        raise TypeError("if reuse_MRQuery is 'new' "
              'then "topdirs" must be given')

    filenames = pathnames.PathNames(topdatadir, basename, topdirs,
                      excluded_dirs)
    if reuse_MRQuery in ('update', 'as_is') and \
              not os.path.isfile(filenames.savefile):
        raise TypeError("\"reuse_MRQuery\" is 'update' "
               "or 'as_is' but no savefile was found")

    if reuse_MRQuery == 'new':
        mrq = MRQuery(topdatadir, basename, topdirs, excluded_dirs)
        mrq.check_invariants()
    elif reuse_MRQuery == 'update':
        mrq = Retrieve_MRQuery(topdatadir, basename)
        if topdirs is not None:
            topdirs = pathnames.prepare_topdirs(topdirs)
        if topdirs is not None and set(topdirs) != set(self.topdirs):
            mrq.replace_topdirs(topdirs)
        if excluded_dirs is not None and \
                     set(excluded_dirs) != set(mrq.filenames.excluded_dirs):
            mrq.unexclude_dirs(excluded_dirs)
        mrq.update()
    elif reuse_MRQuery == 'as_is':
        mrq = Retrieve_MRQuery(topdatadir, basename)
    return mrq

def Do_MRQuery(topdatadir, basename, reuse_MRQuery,
                topdirs=None, excluded_dirs=None):
    """Create or update an instance of MRQuery"""
    if reuse_MRQuery not in ('new', 'update'):
        raise ValueError("reuse_MRQuery must be 'new' or 'update'")
    mrq = FetchOrCreate_MRQuery(topdatadir, basename, reuse_MRQuery, topdirs,
                          excluded_dirs)
    mrq.Store()

#>=== Misc =========================================================

def SortlistInfo(topdatadir, basename):
    mrq = Retrieve_MRQuery(topdatadir, basename)
    bins = 101 * [0]
    for score, r, c in mrq.sortlist:
        bins[int(100*score)] += 1
    return len(mrq.sortlist), bins
