
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os
import constants
from LN import LN, MLLN

class PathNameError(StandardError):
    pass

def splitall(aname):
    """Find components of a pathname"""
    components = aname.split('/')
    if components[0] == '':
        components[0] = '/'
    for i in reversed(range(len(components))):
        if components[i] == '':
             del components[i]
    return components

def isunder(dir1, dir2):
    """Is dir1 under dir2?"""
    dir1 = os.path.abspath(dir1)
    dir2 = os.path.abspath(dir2)
    if dir1.startswith(dir2) and \
            (len(dir1) == len(dir2) or dir1[len(dir2)] == '/'):
        return True
    return False

def gcdir(names, n):
    """Removes from each name a prefix which is a common directory for the names.

    Leaves n common directories on each name.
    Assumes all names are absolute pathnames and names of files.
    "gc" means "greatest common".
    """
    prefix = os.path.commonprefix(names)
    i = prefix.rindex('/')
    prefix = prefix[:i]
    components = splitall(prefix)
    components = components[-n:]
    prefix = os.path.join(*components)
    out = list()
    for name in names:
        short = name[i:]
        if len(short) == 0:
            out.append('/')
        else:
            out.append(prefix + name[i:])
    return out

def commondir(dirs):
    """Finds the lowest common directory for a list of directories.

    Assumes all names are absolute pathnames and names of directories.
    The absolute pathname of a directory does not have a "/" at the end.
    """
    if not isinstance(dirs, (list, tuple)):
        raise TypeError('argument must be a list or tuple')
    prefix = os.path.commonprefix(dirs)
    if prefix == '/':
        return prefix
    if prefix.endswith('/'):
        return prefix[:-1]
    if prefix == min(dirs):
        return prefix
    i = prefix.rindex('/')
    return prefix[:i]

def makedir(name):
    """Create a directory is it does not already exist"""
    name = os.path.abspath(name)
    if os.path.isdir(name):
        return
    if os.path.exists(name) and not os.path.isdir(name):
         raise IOError('file exists but is not a directory: %s' % name)
    os.mkdir(name)

def datadirname(topdatadir, basename):
    """Create the name of the data directory where the "pkl" files are kept"""
    return os.path.join(topdatadir, basename + '_data')

def Xtextfile(topdatadir, basename):
    """Create an Xtext filename. See "html.py". """
    datadir = datadirname(topdatadir, basename)
    return os.path.join(self.datadir, constants.XTEXT)

def savefilename(topdatadir, basename):
    """Create a savefile name"""
    topdatadir = os.path.abspath(topdatadir)
    datadir = os.path.join(topdatadir, basename + '_data')
    savefile = os.path.join(datadir, constants.MRQPKL)
    return topdatadir, basename, savefile

class PathNames(object):
    """Class that creates some useful pathnames"""
    def __init__(self, topdatadir, basename, topdirs=None, excluded_dirs=None):
        self.MRQPKL = constants.MRQPKL
        self.MRQPKLBAK = constants.MRQPKLBAK
        self.XTEXT = constants.XTEXT
        self.topdatadir = os.path.abspath(topdatadir)
        if not os.path.exists(self.topdatadir):
            raise IOError('topdatadir does not exist:"%s"' % self.topdatadir)
        if not os.path.isdir(self.topdatadir):
            raise IOError('topdatadir exists but is not a directory')
        self.basename = basename
        if topdirs is None:
            self.topdirs = None
        else:
            self.topdirs = prepare_topdirs(topdirs)
        if excluded_dirs is None:
            self.excluded_dirs = None
        else:
            self.excluded_dirs = prepare_excludes(excluded_dirs)

        self.datadir = os.path.join(self.topdatadir, self.basename + '_data')
        self.htmldir = os.path.join(self.topdatadir, self.basename + '_html')
        self.thumbdir = os.path.join(self.topdatadir, self.basename + '_thumb')
        self.savefile = os.path.join(self.datadir, self.MRQPKL)
        self.bakfile = os.path.join(self.datadir, self.MRQPKLBAK)
        self.Xtextfile = os.path.join(self.datadir, self.XTEXT)

    def makedirs(self):
        makedir(self.datadir)
        makedir(self.htmldir)
        makedir(self.thumbdir)

def prepare_excludes(excluded_dirs):
    """Normalize anmes of excluded directories"""
    if excluded_dirs is None:
        return ()
    if isinstance(excluded_dirs, str):
        excludes = (excluded_dirs,)
    excludes = set()
    for name in excluded_dirs:
        name = os.path.abspath(name)
        if not os.path.isdir(name):
            LN(name)
            raise IOError('a name in "excluded_dirs" is not the name of a directory')
        excludes.add(name)
    return excludes

def prepare_topdirs(topdirs):
    """Normalize names of the topdirs"""
    if isinstance(topdirs, str):
        topdirs = (topdirs,)
    topdirs2 = list()
    for topdir in topdirs:
        topdir = os.path.abspath(topdir)
        if topdir in topdirs2:
            continue
        if not os.path.exists(topdir):
            raise IOError('topdir does not exist: %s' % topdir)
        else:
            if not os.path.isdir(topdir):
                raise IOError(
                  'topdir exists but is not a directory: %s' % topdir)
        topdirs2.append(topdir)
    for topdir1 in topdirs2:
        for topdir2 in topdirs2:
            if topdir1 != topdir2 and isunder(topdir1, topdir2):
                raise TypeError('a topdir cannot be a subdirectory '
                  'of another topdir:\n  "%s"\n  "%s"' % (topdir1, topdir2))
    return tuple(topdirs2)
