
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, math, time, warnings, textwrap
import constants, pathnames, utilities, mrquery
import haar, weights, scorec
import numarray
from numarray.numerictypes import *
from LN import LN, MLLN

USE_C = True

Wts = weights.weights(constants.HAARSIZE)

#>=== Block ===================================================

"""We need to compare each pair of different images with each other. As an
example, I will use eight images. The pairs that need to be compared are
below the diagonal in the figure below. They are marked with a '*'. The
diagonal elements are '@'s.

The obvious thing to do is to process one row at a time. But this will cause
disk thrashing if the entire MRQuery instance cannot be held in memory at
one time. So we divide the triangle into blocks. In the example, the blocks
will have 3 rows and 4 columns.

The progranm is designed so it can be interrupted witout losing more than
two minutes of work.

 row
     -------
  0 | @   o | o   o   o   o   o   o
  1 | *   @ | o   o   o   o   o   o
  2 | *   * | @   o   o   o   o   o
     ------- ------- ---
  3 | *   *   *   @ | o | o   o   o  <- For this block, rstart = 3, cstart = 4,
  4 | *   *   *   * | @ | o   o   o     rsize = 3 and csize = 1. This block is
  5 | *   *   *   * | * | @   o   o     at the end of a row because
     --------------- --- -------        rstart + rsize - 1 == cstart + csize.
  6 | *   *   *   * | *   *   @ | o
  7 | *   *   *   * | *   *   * | @
     --------------- -----------
      0   1   2   3   4   5   6   7 col
"""
class Block(object):
    """Contains data needed to process a sequence of blocks of scores"""
    def __init__(self, rows):
        """Creates the block "before" the first block.

        Call "next" to get the first block. Call "next" to get each
        successive block.

        Note: There is actually only one instance of Block which is updated
        each time "next" is called.
        """
        self.rowblocksize = constants.RBLOCKSIZE  # Maximum number of rows
        self.colblocksize = constants.CBLOCKSIZE  # and columns permitted
                                                  # in the block.
        self.rows = rows    # Total number of images.
        self.rstart = None  # Starting row for this block.
        self.rsize = None   # Rows in this block.
        self.cstart = None  # Starting column for this block.
        self.csize = None   # Columns in this block.
        self.pairsdone = 0  # Number of pairs that have been scored.
        self.counter = 0    # The number of completed blocks.

    def at_end_of_row(self):
        """Is a block at the end of a row of blocks?"""
        return self.rstart + self.rsize - 1 == self.cstart + self.csize

    def chi(self, r):
        """Find the column after the last column in this block"""
        if self.rstart is None:
            return None
        return max(self.cstart, min(self.cstart + self.colblocksize, r))

    def do_pairshere(self):
        """Find the number of pairs of scores that will be processed in this block.

        This number is approximate.
        """
        if self.rstart is None:
            return None
        pairs = 0
        for r in range(self.rstart, self.rstart + self.rsize):
            pairs += self.chi(r) - self.cstart
        return pairs

    def find_totalpairs(self):
        """Find the total number of pairs of scores.

        This number is approximate.
        """
        if self.rstart is None:
            return None
        return self.rows * (self.rows - 1) // 2

    def alldone(self):
        """Do any more blocks need to be done?"""
        if self.rstart is None:
            return False
        return self.at_end_of_row() and self.rstart + self.rsize == self.rows

    def first_uncompleted_imnum(self):
        """Find the first imnum that has not been completely scored.

        Use this in ranges, etc.
        """
        if self.at_end_of_row():
            return self.rstart + self.rsize
        else:
            return self.rstart

    def last_completed_imnum(self):
        """Find the largest imnum that has been completely scored"""
        return self.first_uncompleted_imnum() - 1

    def __repr__(self):
        return repr( (self.rows, self.rstart, self.cstart, self.rsize,
                      self.csize) )

    def __str__(self):
        template = """\
            <Block: rows: %s,
                rstart: %s, cstart: %s, rsize: %s, csize: %s,
                pairshere: %s"""
        template = textwrap.dedent(template)
        return template % (self.rows, self.rstart, self.cstart,
                 self.rsize, self.csize, self.pairshere())

    def next(self):
        """Modifies a block to produce the next block.

        A Block object may be pickled and stored to disk. By the time the
        Block is read from disk, more rows may have been added to the data.
        Therefore, before calling next for the first time, "self.rows" must
        be set. It will usually be "haar_count" from MRQuery.

        The block created by Block() is the block "before" the first block.
        Here is how "next" is used:
            self.mrq.block.rows = self.mrq.haar_count
            while self.mrq.block.next():
                ...
        """
        # Between calls to "next", this block may be kept in a file so
        # "constants.py may have been changed.  I think it is always safe to
        # change "self.colblocksize".
        self.colblocksize = constants.CBLOCKSIZE
        if self.rstart is None:
            # First block.
            rstart = 0
            rsize = min(self.rows, self.rowblocksize)
            cstart = 0
            csize = min(rsize - 1, self.colblocksize)
        elif self.at_end_of_row():
            # This block is currently at the end of a row.
            if self.alldone():
                # Completely finished.
                return False
            else:
                # Between calls to "next", this block may be kept in a file
                # so "constants.py may have been changed.  If RBLOCKSIZE is
                # modified, it is safe (?) to change it here.
                self.rowblocksize = constants.RBLOCKSIZE
                rstart = self.rstart + self.rsize
                rsize =  min(self.rows - rstart, self.rowblocksize)
                cstart = 0
                csize = min(rstart + rsize - 1, self.colblocksize)
        else:
            # This block is not at the end of a row of blocks.
            rstart = self.rstart
            rsize = self.rsize
            cstart = self.cstart + self.colblocksize
            csize = min(rstart + rsize - 1 - cstart, self.colblocksize)

        # Modify this block to get the next one.
        self.rstart = rstart
        self.rsize = rsize
        self.cstart = cstart
        self.csize = csize
        self.pairshere = self.do_pairshere()
        self.pairsdone += self.pairshere
        self.counter += 1
        return True

#>=== Scorearr =======================================================

class Scorearr(object):
    """Finds a score for each pair of images"""
    def __init__(self, topdatadir, basename, reuse_sort, usewts=True):
        # "reuse_sort" can be:
        #  'new'      Start Scorearr using the output of Haar
        #  'continue' Continue doing Scorearr.
        if reuse_sort not in ('new', 'continue'):
            raise TypeError("\"reuse_sort\" must be 'new' or 'continue'")

        filenames = pathnames.PathNames(topdatadir, basename)
        self.mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
        self.mrq.ddl_sortlist = list()  # Save space.
        # The algorithms in Scorearr do not work unless self.mrq is
        # cleaned.
        self.mrq.clean()
        self.reuse_sort = reuse_sort
        # See "weights.py".
        self.mrq.usewts = usewts
        self.cols = self.mrq.imnum_count
        self.mrq.block.rows = self.mrq.haar_count
        if self.reuse_sort == 'new':
            self.mrq.resetScorearr(usewts)
        else:
            self.mrq.update_sortlist()
            if self.mrq.block.alldone():
                print 'All the images have already been scored. Nothing to do.'
                return
        self.doloop()

    def doloop(self):
        """Main loop in Scorearr"""
        print 'Starting scorearr loop.'
        # Watch the clock closely.
        t0 = starting_time = savetime = time.time()
        while self.mrq.block.next():
            t = time.time()            
            Ctimesum = 0.0
            block = self.mrq.block
            scorearr = numarray.zeros( (block.rsize * block.csize, 3), Float64)
            if USE_C:
                bad = self.mrq.bad_imnums()
                tt = time.time()
                ctr = scorec.inner_Scorearr(block.rstart, block.cstart,
                     block.rsize, constants.CBLOCKSIZE, self.mrq.usewts,
                     constants.SORTLIST_CUTOFF_SCORE, bad,
                     self.mrq.bwbits, self.mrq.Darrs, Wts.wtsarr,
                     Wts.wtsarr0, scorearr)
                Ctimesum += time.time() - tt
            else:
                ctr = self.inner_Scorearr(block, scorearr)

            # Sort.
            scorearr = scorearr[:ctr, :]
            for r in range(scorearr.shape[0]):
                score, rimnum, cimnum = scorearr[r]
                self.mrq.sortlist.append( (score, int(rimnum), int(cimnum)) )
            self.mrq.sortlist.sort(reverse=True)
            # Note: Here is how to do the sort on a numarray. If the Python
            # sort is too slow, do something like:
            # # "-" reverses the sort.
            # ind = numarray.argsort(-self.mrq.sortarr[:, 0])
            # for i in (0,1,2):
            #     self.mrq.sortarr[:, i] = self.mrq.sortarr[ind, i]
            self.mrq.sortlist = \
                    self.mrq.sortlist[:constants.SORTSCORES_QUOTA]
            if time.time() - savetime >= constants.SAVETIME:
                self.mrq.Store()
                savetime = time.time()

            # Print information about the block.
            current_time = time.time()
            total = block.find_totalpairs()
            if total > 0:
                percent = 100*(float(block.pairsdone)/total)
            else:
                percent = 0.0
            template = 'Block %i, (%s, %s), %i of %i done, %5.2f%%.'
            print template % (block.counter, block.rstart, block.cstart,
                     block.pairsdone, total, percent)
            print '  %4.2f secs, %4.2f secs total' % (current_time - t,
                      current_time - starting_time)
        if block.counter > 0:
            self.mrq.Store()

    def getweight(self, s, layer):
        """Find the weighted score for a set of rc's"""
        wt = 0.0
        for rc in s:
            # Unpack rc into a row and a column. Grep "*.py" for "rcarr".
            r, c = utilities.div(rc, constants.HAARSIZE)
            # In the paper, the weights are subtracted here. If I ever put
            # in an option to use avg_intensity, it will be added. I doubt
            # if I will ever do this since the algroithm works well as it is.
            wt += Wts.wtsarr[layer, abs(r), abs(c)]
        return wt

    def getweight_bw(self, s):
        """Find the weighted score for a set of rc's for black and white images."""
        wt = 0.0
        for rc in s:
            r, c = utilities.div(rc, constants.HAARSIZE)
            # In the paper, the weights are subtracted here. If I ever put
            # in an option to use avg_intensity, it will be subtracted.
            wt += Wts.wtsarr0[abs(r), abs(c)]
        return wt

    def compare_pair(self, r, c):
        """Compare two images.

        See L2norm.txt for the math.
        """
        # If either image is b&w compare only using layer 0.
        if self.mrq.bwbits[r] == True or \
                      self.mrq.bwbits[c] == True:
            rset = set(self.mrq.Darrs[0, r])
            cset = set(self.mrq.Darrs[0, c])
            # This is the only place where the signs in rcarr are used.
            intersection = rset & cset
            # Raw score.
            if self.mrq.usewts:
                Iscore = self.getweight_bw(intersection)
            else:
                Iscore = len(intersection)
            if self.mrq.usewts:
                Iscore = self.getweight_bw(intersection)
                rwt = self.getweight_bw(rset)
                cwt = self.getweight_bw(cset)
            else:
                Iscore = len(intersection)
                rwt = len(rset)
                cwt = len(cset)
            # Normalize using L2 norm.
            L2score = Iscore / math.sqrt(rwt * cwt)
            return L2score
        else:
            Isum = 0.0
            Rsum = 0.0
            Csum = 0.0
            for layer in range(constants.LAYERS):
                rset = set(self.mrq.Darrs[layer, r])
                cset = set(self.mrq.Darrs[layer, c])
                intersection = rset & cset
                if self.mrq.usewts:
                    Isum += self.getweight(intersection, layer)
                    Rsum += self.getweight(rset, layer)
                    Csum += self.getweight(cset, layer)
                else:
                    Isum += len(intersection)
                    Rsum += len(rset)
                    Csum += len(cset)
            # Raw score.
            Iscore = float(Isum)
            # Normalize using L2 norm.
            L2score = Iscore / math.sqrt(Rsum * Csum)
            return L2score

    def inner_Scorearr(self, block, arrout):
        """Main loop for Scorearr.

        Runs in O(rows*cols) time. Scorearr runs in time quadratic in the
        number of images.
        """
        ctr = 0
        for r in range(block.rstart, block.rstart + block.rsize):
            if r not in self.mrq.imnum2hash:
                continue
            chi = max(block.cstart, min(block.cstart + block.colblocksize, r))
            for c in range(block.cstart, chi):
                if c not in self.mrq.imnum2hash:
                    continue
                score = self.compare_pair(r, c)
                if score >= constants.SORTLIST_CUTOFF_SCORE:
                    arrout[ctr, :] = (score, r, c)
                    ctr += 1
        return ctr

#>=== PrintScores ============================================

def PrintScores(topdatadir, basename, which_sortlist='sortlist',
         prefixdirs=None, quota=constants.PRINTSCORE_QUOTA):
    """Prints the scores to stdout"""
    if prefixdirs is None:
        prefixdirs = 1
    elif prefixdirs == 'all':
        prefixdirs = sys.maxint
    mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
    print 'PrintScores:', len(mrq.hash2ImData), 'images.'
    if not mrq.haar_finished():
        warning.warn('haar not finished')
    if not mrq.block.alldone():
        warnings.warn('sorting not completed')
    sortlist = getattr(mrq, which_sortlist)
    name_count = mrq.hash_count()
    totalpairs = name_count*(name_count-1)//2

    print basename
    printed = min(quota, len(sortlist))
    template = """%i different images, %i pairs printed, %i possible, %s has %i."""
    print template % (name_count, printed, totalpairs, which_sortlist,
                              len(sortlist))

    # Print the duplicate images.
    print 'Sets of identical images:'
    for hashcode in mrq.hash2ImData:
        ID = mrq.hash2ImData[hashcode]
        if len(ID) > 1:
            names = ID.tonames()
            names = pathnames.gcdir(names, prefixdirs)
            for name in names:
                print name
            print

    print 'L2score'
    if len(sortlist) == 0:
        return
    template = '%6.2f %-20s %s'
    ctr = 0
    if len(sortlist[0]) == 3:
        for L2score, imnum1, imnum2 in sortlist:
            if imnum1 == imnum2:
                raise Exception('BUG')
            hashcode1 = mrq.imnum2hash[imnum1]
            name1 = mrq.hash2ImData[hashcode1][0]
            hashcode2 = mrq.imnum2hash[imnum2]
            name2 = mrq.hash2ImData[hashcode][0]
            name1, name2 = pathnames.gcdir((name1, name2), prefixdirs)
            print template % (L2score, name1, name2)
            ctr += 1
            if ctr > quota:
                break
    else:
        for score, N1, N2, ID1, ID2 in sortlist:
            name1 = N1.as_str()
            name2 = N2.as_str()
            name1, name2 = pathnames.gcdir((name1, name2), prefixdirs)
            print template % (score, name1, name2)
            ctr += 1
            if ctr > quota:
                break
